#include <msp/gl/blend.h>
#include <msp/gl/tests.h>
#include "layout.h"
#include "track.h"
#include "view.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

View3D::View3D(Layout3D &l, unsigned w, unsigned h):
	layout(l),
	width(w),
	height(h),
	pipeline(w, h),
	sky(layout.get_catalogue()),
	shadow(4096, layout.get_scene(), layout.get_sun()),
	ambient_occlusion(w, h, 100)
{
	pipeline.set_hdr(true);
	pipeline.set_camera(&camera);
	pipeline.add_renderable(sky);
	pipeline.add_renderable_for_pass(shadow, 0);
	pipeline.add_renderable_for_pass(layout.get_scene(), "translucent");

	GL::Pipeline::Pass *pass = &pipeline.add_pass("sky");
	pass->set_lighting(&layout.get_lighting());

	pass = &pipeline.add_pass(0);
	pass->set_lighting(&layout.get_lighting());
	pass->set_depth_test(&GL::DepthTest::lequal());

	pass = &pipeline.add_pass("translucent");
	pass->set_lighting(&layout.get_lighting());
	pass->set_depth_test(&GL::DepthTest::lequal());
	pass->set_blend(&GL::Blend::alpha());

	pipeline.add_postprocessor(ambient_occlusion);
	colorcurve.set_srgb();
	pipeline.add_postprocessor(colorcurve);

	update_shadow_area();
	layout.get_layout().signal_object_added.connect(sigc::hide(sigc::mem_fun(this, &View3D::update_shadow_area)));

	camera.set_up_direction(GL::Vector3(0, 0, 1));
	// Y+, 60° down
	camera.set_look_direction(GL::Vector3(0, 0.5, -0.866));
	camera.set_aspect(float(width)/height);

	view_all();
}

Ray View3D::create_ray(int x, int y) const
{
	return create_ray(x*2.0f/width-1.0f, y*2.0f/height-1.0f);
}

Ray View3D::create_ray(float x, float y) const
{
	const GL::Vector3 &start = camera.get_position();
	GL::Vector4 ray = camera.unproject(GL::Vector4(x, y, 0, 0));
	return Ray(start, ray.slice<3>(0));
}

void View3D::compute_bounds(Vector &minp, Vector &maxp)
{
	const set<Object *> &objects = layout.get_layout().get_all<Object>();
	Geometry::BoundingBox<float, 3> bbox;
	for(set<Object *>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		bbox = bbox|(*i)->get_bounding_box();

	minp = bbox.get_minimum_point();
	maxp = bbox.get_maximum_point();
}

void View3D::view_all(bool tight)
{
	Vector minp, maxp;
	compute_bounds(minp, maxp);

	float t = tan(camera.get_field_of_view()/2.0f)*2.0f;
	float size = max((maxp.y-minp.y+0.1), (maxp.x-minp.x+0.1)/camera.get_aspect());
	float dist = size/t;
	if(!tight)
		dist += sin(camera.get_field_of_view()/2.0f)*size;
	GL::Vector3 center = (minp+maxp)/2.0f;
	const GL::Vector3 &look = camera.get_look_direction();
	camera.set_position(center-look*dist);
	camera.set_depth_clip(dist*0.02, dist*50);
}

void View3D::render()
{
	pipeline.render();
}

void View3D::update_shadow_area()
{
	Vector minp, maxp;
	compute_bounds(minp, maxp);
	shadow.set_target((minp+maxp)/2.0f, max((maxp-minp).norm()/2.0f, 0.1f));
}

} // namespace R2C2
