#include <cmath>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "libr2c2/catalogue.h"
#include "libr2c2/trackappearance.h"
#include "libr2c2/track.h"
#include "libr2c2/tracktype.h"
#include "svgexporter.h"

using namespace std;
using namespace Msp;
using namespace R2C2;

SvgExporter::SvgExporter(const Layout &l):
	layout(l)
{ }

void SvgExporter::save(const string &fn)
{
	xmlpp::Document *doc = new xmlpp::Document;
	root = doc->create_root_node("svg", "http://www.w3.org/2000/svg");

	xmlpp::Element *style = root->add_child("style");
	style->set_attribute("type", "text/css");
	styles.push_back(".rail { fill: none; stroke: #000000; }");
	styles.push_back(".border { fill: none; stroke: #808080; }");
	styles.push_back(".artnr { text-anchor: middle; font-family: sans; }");

	defs = root->add_child("defs");

	Vector minp;
	Vector maxp;
	const set<Track *> &tracks = layout.get_all<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
	{
		save_track(**i);

		float gauge = (*i)->get_type().get_gauge();
		unsigned n_endpoints = (*i)->get_type().get_endpoints().size();
		for(unsigned j=0; j<n_endpoints; ++j)
		{
			Vector pos = (*i)->get_snap_node(j).position;
			if(i==tracks.begin() && j==0)
				minp = maxp = pos;
			else
			{
				minp.x = min(minp.x, pos.x-gauge*3);
				minp.y = min(minp.y, pos.y-gauge*3);
				maxp.x = max(maxp.x, pos.x+gauge*3);
				maxp.y = max(maxp.y, pos.y+gauge*3);
			}
		}
	}

	style->set_child_text(join(styles.begin(), styles.end(), "\n"));

	root->set_attribute("viewBox", format("%.3f %.3f %.3f %.3f",
		minp.x*1000, -maxp.y*1000, (maxp.x-minp.x)*1000, (maxp.y-minp.y)*1000));

	doc->write_to_file_formatted(fn);
}

string SvgExporter::create_appearance(const TrackAppearance &appearance)
{
	float gauge = appearance.get_gauge();
	string key = format("a%.0f", gauge*1000);

	if(!appearances_created.count(&appearance))
	{
		float rail_width = appearance.get_rail_profile().get_width();

		styles.push_back(format(".%s .rail { stroke-width: %.3f }", key, rail_width*1000));
		styles.push_back(format(".%s .border { stroke-width: %.3f }", key, rail_width*500));
		styles.push_back(format(".%s .artnr { font-size: %.3f }", key, gauge*750));

		appearances_created.insert(&appearance);
	}

	return key;
}

string SvgExporter::create_point_data(const Vector &p)
{
	return format("%.3f %.3f", p.x*1000, -p.y*1000);
}

string SvgExporter::create_line_data(const Vector &s, const Vector &e)
{
	return format("M %s L %s", create_point_data(s), create_point_data(e));
}

string SvgExporter::create_curve_data(const Vector &s, const Vector &st, const Vector &et, const Vector &e)
{
	return format("M %s C %s %s %s", create_point_data(s),
		create_point_data(s+st), create_point_data(e-et), create_point_data(e));
}

string SvgExporter::create_track_type(const TrackType &type)
{
	const ArticleNumber &art_nr = type.get_article_number();
	string key = format("t%s", art_nr.str());

	if(!tracks_created.count(&type))
	{
		const TrackAppearance &appearance = type.get_appearance();
		string appearance_key = create_appearance(appearance);

		xmlpp::Element *group = defs->add_child("g");
		group->set_attribute("id", key);
		group->set_attribute("class", appearance_key);

		float gauge = appearance.get_gauge();
		float rail_width = appearance.get_rail_profile().get_width();

		const vector<TrackType::Endpoint> &endpoints = type.get_endpoints();
		for(vector<TrackType::Endpoint>::const_iterator i=endpoints.begin(); i!=endpoints.end(); ++i)
		{
			xmlpp::Element *path = group->add_child("path");
			path->set_attribute("class", "border");

			Vector delta = rotated_vector(Vector(0, gauge, 0), i->dir);

			path->set_attribute("d", create_line_data(i->pos+delta, i->pos-delta));
		}

		const vector<TrackPart> &parts = type.get_parts();
		for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
		{
			xmlpp::Element *path = group->add_child("path");
			path->set_attribute("class", "rail");

			string data;

			OrientedPoint start = i->get_point(0);
			OrientedPoint end = i->get_point(i->get_length());
			if(i->is_curved())
			{
				Vector delta1 = rotated_vector(Vector(0, (gauge+rail_width)*0.5, 0), start.rotation);
				Vector delta2 = rotated_vector(Vector(0, (gauge+rail_width)*0.5, 0), end.rotation);
				// Largely an educated guess, but seems to be accurate enough
				float clen = i->get_length()/2.9;
				Vector ctrl1 = rotated_vector(Vector(clen, 0, 0), start.rotation);
				Vector ctrl2 = rotated_vector(Vector(clen, 0, 0), end.rotation);

				data = create_curve_data(start.position+delta1, ctrl1, ctrl2, end.position+delta2);
				data += ' ';
				data += create_curve_data(start.position-delta1, ctrl1, ctrl2, end.position-delta2);
			}
			else
			{
				Vector delta = rotated_vector(Vector(0, (gauge+rail_width)*0.5, 0), start.rotation);

				data = create_line_data(start.position+delta, end.position+delta);
				data += ' ';
				data += create_line_data(start.position-delta ,end.position-delta);
			}

			path->set_attribute("d", data);
		}

		OrientedPoint label_pt = parts.front().get_point(parts.front().get_length()/2);

		label_pt.rotation.wrap_with_base(-Angle::quarter_turn());
		if(label_pt.rotation>Angle::quarter_turn())
			label_pt.rotation -= Angle::half_turn();

		label_pt.position += rotated_vector(Vector(0, -gauge*0.25, 0), label_pt.rotation);

		xmlpp::Element *text = group->add_child("text");
		text->set_attribute("class", "artnr");
		text->set_attribute("transform", format("translate(%s) rotate(%.3f)",
			create_point_data(label_pt.position), -label_pt.rotation.degrees()));
		text->set_child_text(type.get_article_number().str());

		tracks_created.insert(&type);
	}

	return key;
}

void SvgExporter::save_track(const Track &track)
{
	string type_key = create_track_type(track.get_type());
	xmlpp::Element *use = root->add_child("use");

	const Vector &pos = track.get_position();
	const Angle &rot = track.get_rotation();
	string transform = format("translate(%s) rotate(%.3f)", create_point_data(pos), -rot.degrees());
	use->set_attribute("transform", transform);

	use->set_attribute("xlink:href", "#"+type_key);
}
