#include <algorithm>
#include <msp/core/maputils.h>
#include <msp/core/raii.h>
#include <msp/core/refptr.h>
#include <msp/datafile/parser.h>
#include <msp/datafile/writer.h>
#include <msp/io/print.h>
#include <msp/time/utils.h>
#include "beamgate.h"
#include "block.h"
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "route.h"
#include "signal.h"
#include "signaltype.h"
#include "terrain.h"
#include "track.h"
#include "trackcircuit.h"
#include "tracktype.h"
#include "train.h"
#include "vehicle.h"
#include "vehicletype.h"
#include "zone.h"

using namespace std;
using namespace Msp;

namespace {

bool zone_order(const R2C2::Zone *z1, const R2C2::Zone *z2)
{
	return z1->get_number()<z2->get_number();
}

}


namespace R2C2 {

Layout::Layout(Catalogue &c, Driver *d):
	catalogue(c),
	driver(d),
	next_turnout_addr(0x800),
	emitting_block_reserved(false)
{
	clock.set_rate(60);
}

Layout::~Layout()
{
	delete driver;
	driver = 0;

	while(!trains.empty())
		delete trains.begin()->second;
	track_chains.del<Route>();
	track_chains.del<Zone>();
	objects.del<Signal>();
	objects.del<Track>();
	track_chains.del<Block>();
}

Driver &Layout::get_driver() const
{
	if(!driver)
		throw logic_error("!driver");
	return *driver;
}

void Layout::add(Object &o)
{
	if(objects.insert(o))
	{
		try
		{
			signal_object_added.emit(o);
		}
		catch(...)
		{
			objects.erase(o);
			throw;
		}
	}
}

void Layout::add(Track &t)
{
	if(objects.insert(t))
	{
		try
		{
			// Blocks must be recreated first
			create_blocks();
			signal_object_added.emit(t);
		}
		catch(...)
		{
			objects.erase(t);
			create_blocks();
			throw;
		}
	}
}

void Layout::add(TrackChain &g)
{
	if(track_chains.insert(g))
	{
		try
		{
			signal_track_chain_added.emit(g);
		}
		catch(...)
		{
			track_chains.erase(g);
			throw;
		}
	}
}

void Layout::add(Block &b)
{
	if(track_chains.insert(b))
	{
		sigc::connection conn = b.signal_reserved.connect(sigc::bind<0>(sigc::mem_fun(this, &Layout::block_reserved), sigc::ref(b)));
		try
		{
			signal_track_chain_added.emit(b);
		}
		catch(...)
		{
			track_chains.erase(b);
			conn.disconnect();
			throw;
		}
	}
}

void Layout::add(Sensor &s)
{
	if(sensors.insert(s))
	{
		s.signal_state_changed.connect(sigc::bind<0>(sigc::mem_fun(this, &Layout::sensor_state_changed), sigc::ref(s)));
		s.signal_state_changed.connect(sigc::bind<0>(signal_sensor_state_changed, sigc::ref(s)));
	}
}

void Layout::add(BeamGate &g)
{
	add(static_cast<Object &>(g));
	add(static_cast<Sensor &>(g));
}

template<>
const set<Object *> &Layout::get_all<Object>() const
{
	return objects.get();
}

template<>
const set<Track *> &Layout::get_all<Track>() const
{
	return objects.get<Track>();
}

template<>
const set<Signal *> &Layout::get_all<Signal>() const
{
	return objects.get<Signal>();
}

template<>
const set<Vehicle *> &Layout::get_all<Vehicle>() const
{
	return objects.get<Vehicle>();
}

template<>
const set<Terrain *> &Layout::get_all<Terrain>() const
{
	return objects.get<Terrain>();
}

template<>
const set<TrackChain *> &Layout::get_all<TrackChain>() const
{
	return track_chains.get();
}

template<>
const set<Block *> &Layout::get_all<Block>() const
{
	return track_chains.get<Block>();
}

template<>
const set<Route *> &Layout::get_all<Route>() const
{
	return track_chains.get<Route>();
}

template<>
const set<Zone *> &Layout::get_all<Zone>() const
{
	return track_chains.get<Zone>();
}

template<>
const set<Sensor *> &Layout::get_all<Sensor>() const
{
	return sensors.get<Sensor>();
}

template<>
const set<TrackCircuit *> &Layout::get_all<TrackCircuit>() const
{
	return sensors.get<TrackCircuit>();
}

template<>
const set<BeamGate *> &Layout::get_all<BeamGate>() const
{
	return sensors.get<BeamGate>();
}

template<typename T>
T *Layout::pick(const Ray &ray)
{
	const set<Object *> &objs = objects.get();
	T *closest = 0;
	float distance = -1;
	for(set<Object *>::const_iterator i=objs.begin(); i!=objs.end(); ++i)
		if(T *t = dynamic_cast<T *>(*i))
		{
			float d = -1;
			if(t->collide_ray(ray, &d))
				if(!closest || d<distance)
				{
					closest = t;
					distance = d;
				}
		}

	return closest;
}

template Object *Layout::pick<Object>(const Ray &);
template Track *Layout::pick<Track>(const Ray &);
template Signal *Layout::pick<Signal>(const Ray &);
template Vehicle *Layout::pick<Vehicle>(const Ray &);

void Layout::remove(Object &o)
{
	if(objects.erase(o))
		signal_object_removed.emit(o);
}

void Layout::remove(Track &t)
{
	if(objects.erase(t))
	{
		// Blocks must be recreated first
		create_blocks(t);
		signal_object_removed.emit(t);
	}
}

void Layout::remove(TrackChain &g)
{
	if(track_chains.erase(g))
		signal_track_chain_removed.emit(g);
}

void Layout::remove(Sensor &s)
{
	sensors.erase(s);
}

void Layout::remove(BeamGate &g)
{
	remove(static_cast<Object &>(g));
	remove(static_cast<Sensor &>(g));
}

unsigned Layout::allocate_turnout_address()
{
	set<unsigned> used_addrs;
	const set<Track *> &tracks = objects.get<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_type().is_turnout())
			used_addrs.insert((*i)->get_turnout_address());

	unsigned result = next_turnout_addr;
	while(used_addrs.count(result))
		++result;
	next_turnout_addr = result+1;

	return result;
}

Block &Layout::get_block(unsigned id) const
{
	const set<Block *> &blocks = track_chains.get<Block>();
	for(set<Block *>::const_iterator i=blocks.begin(); i!=blocks.end(); ++i)
		if((*i)->get_id()==id)
			return **i;

	throw key_error(id);
}

void Layout::create_blocks()
{
	set<Track *> loose_tracks = objects.get<Track>();
	const set<Block *> *blocks = &track_chains.get<Block>();
	for(set<Block *>::const_iterator i=blocks->begin(); i!=blocks->end(); ++i)
	{
		const set<Track *> &btracks = (*i)->get_tracks();
		for(set<Track *>::const_iterator j=btracks.begin(); j!=btracks.end(); ++j)
			loose_tracks.erase(*j);
	}

	list<Block *> created_blocks;
	while(!loose_tracks.empty())
	{
		Block *block = new Block(*this, **loose_tracks.begin());
		created_blocks.push_back(block);

		const set<Track *> &btracks = block->get_tracks();
		for(set<Track *>::const_iterator i=btracks.begin(); i!=btracks.end(); ++i)
			loose_tracks.erase(*i);
	}

	// The previously obtained set has been invalidated by creating new blocks
	blocks = &track_chains.get<Block>();
	for(list<Block *>::iterator i=created_blocks.begin(); i!=created_blocks.end(); ++i)
		for(set<Block *>::const_iterator j=blocks->begin(); j!=blocks->end(); ++j)
			if(*j!=*i)
				(*i)->check_link(**j);
}

void Layout::create_blocks(Track &track)
{
	/* Must collect the blocks in a set first while all tracks are still
	guaranteed to have blocks and to avoid duplicate deletes */
	set<Block *> del_blocks;

	del_blocks.insert(&track.get_block());

	const vector<Track *> &links = track.get_links();
	for(vector<Track *>::const_iterator i=links.begin(); i!=links.end(); ++i)
		if(*i)
			del_blocks.insert(&(*i)->get_block());

	for(set<Block *>::iterator i=del_blocks.begin(); i!=del_blocks.end(); ++i)
		delete *i;

	create_blocks();
}

Route &Layout::get_route(const string &name) const
{
	const set<Route *> &routes = track_chains.get<Route>();
	for(set<Route *>::const_iterator i=routes.begin(); i!=routes.end(); ++i)
		if((*i)->get_name()==name)
			return **i;
	throw key_error(name);
}

void Layout::update_routes()
{
	const set<Route *> &routes = track_chains.get<Route>();
	for(set<Route *>::iterator i=routes.begin(); i!=routes.end(); ++i)
		(*i)->update_turnouts();
}

Layout::ZoneArray Layout::get_zones(const string &group) const
{
	ZoneArray result;
	const set<Zone *> &zones = track_chains.get<Zone>();
	for(set<Zone *>::const_iterator i=zones.begin(); i!=zones.end(); ++i)
		if((*i)->get_group()==group)
			result.push_back(*i);

	sort(result.begin(), result.end(), zone_order);

	return result;
}

Zone &Layout::get_zone(const string &group, unsigned num) const
{
	const set<Zone *> &zones = track_chains.get<Zone>();
	for(set<Zone *>::const_iterator i=zones.begin(); i!=zones.end(); ++i)
		if((*i)->get_group()==group && (*i)->get_number()==num)
			return **i;

	throw key_error(format("%s %d", group, num));
}

void Layout::add_train(Train &t)
{
	insert_unique(trains, t.get_address(), &t);
	signal_train_added.emit(t);
}

Train &Layout::get_train(unsigned addr) const
{
	return *get_item(trains, addr);
}

void Layout::remove_train(Train &t)
{
	if(trains.erase(t.get_address()))
		signal_train_removed.emit(t);
}

void Layout::tick(float speed)
{
	if(driver)
		driver->tick();

	Time::TimeStamp t = Time::now();
	Time::TimeDelta dt;
	if(last_tick)
		dt = t-last_tick;
	last_tick = t;

	dt *= speed;
	unsigned count = dt/(10*Time::msec)+1;
	dt /= count;
	while(count--)
		step(dt);
}

void Layout::step(const Time::TimeDelta &dt)
{
	if(!driver || !driver->is_halted())
		clock.tick(dt);

	for(set<Sensor *>::iterator i=sensors.get().begin(); i!=sensors.get().end(); ++i)
		(*i)->tick(dt);
	const set<Signal *> &signals = objects.get<Signal>();
	for(set<Signal *>::iterator i=signals.begin(); i!=signals.end(); ++i)
		(*i)->tick(dt);
	for(map<unsigned, Train *>::iterator i=trains.begin(); i!=trains.end(); ++i)
		i->second->tick(dt);
}

void Layout::emergency(Block *block, const string &msg)
{
	if(driver)
		driver->halt(true);
	if(block)
		IO::print("Emergency at %s: %s\n", block->get_name(), msg);
	else
		IO::print("Emergency: %s\n", msg);
	signal_emergency.emit(block, msg);
}

void Layout::save(const string &fn) const
{
	IO::BufferedFile out(fn, IO::M_WRITE);
	DataFile::Writer writer(out);

	if(!base.empty())
		writer.write((DataFile::Statement("base"), base));

	const set<Track *> &tracks = objects.get<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
	{
		DataFile::Statement st("track");
		st.append(catalogue.get_name(&(*i)->get_type()));
		(*i)->save(st.sub);
		writer.write(st);
	}

	const set<Signal *> &signals = objects.get<Signal>();
	for(set<Signal *>::const_iterator i=signals.begin(); i!=signals.end(); ++i)
	{
		DataFile::Statement st("signal");
		st.append(catalogue.get_name(&(*i)->get_type()));
		(*i)->save(st.sub);
		writer.write(st);
	}

	const set<Terrain *> &terrains = objects.get<Terrain>();
	for(set<Terrain *>::const_iterator i=terrains.begin(); i!=terrains.end(); ++i)
	{
		DataFile::Statement st("terrain");
		st.append(catalogue.get_name(&(*i)->get_type()));
		(*i)->save(st.sub);
		writer.write(st);
	}

	const set<BeamGate *> &gates = objects.get<BeamGate>();
	for(set<BeamGate *>::const_iterator i=gates.begin(); i!=gates.end(); ++i)
	{
		DataFile::Statement st("beamgate");
		(*i)->save(st.sub);
		writer.write(st);
	}

	const set<Route *> &routes = track_chains.get<Route>();
	for(set<Route *>::const_iterator i=routes.begin(); i!=routes.end(); ++i)
	{
		if((*i)->is_temporary())
			continue;

		DataFile::Statement st("route");
		(*i)->save(st.sub);
		writer.write(st);
	}

	const set<Zone *> &zones = track_chains.get<Zone>();
	for(set<Zone *>::const_iterator i=zones.begin(); i!=zones.end(); ++i)
	{
		DataFile::Statement st("zone");
		(*i)->save(st.sub);
		writer.write(st);
	}
}

void Layout::save_dynamic(const string &fn) const
{
	IO::BufferedFile out(fn, IO::M_WRITE);
	DataFile::Writer writer(out);

	{
		DataFile::Statement st("clock");
		clock.save(st.sub);
		writer.write(st);
	}

	const set<Track *> &tracks = objects.get<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_type().is_turnout())
		{
			DataFile::Statement st("turnout");
			st.append((*i)->get_turnout_address());
			(*i)->save_dynamic(st.sub);
			writer.write(st);
		}

	for(map<unsigned, Train *>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
	{
		DataFile::Statement st("train");
		st.append(catalogue.get_name(&i->second->get_locomotive_type()));
		st.append(i->second->get_address());
		st.append(i->second->get_protocol());
		i->second->save(st.sub);
		writer.write(st);
	}
}

void Layout::sensor_state_changed(Sensor &sensor, Sensor::State state)
{
	if(state==Sensor::ACTIVE)
	{
		Block *block = sensor.get_block();
		if(block && !block->get_train())
			emergency(block, "Unreserved sensor triggered");
	}
}

void Layout::block_reserved(Block &block, Train *train)
{
	if(emitting_block_reserved)
		block_reserve_queue.push_back(BlockReservation(block, train));
	else
	{
		SetFlag setf(emitting_block_reserved);
		signal_block_reserved.emit(block, train);
		while(!block_reserve_queue.empty())
		{
			BlockReservation br = block_reserve_queue.front();
			block_reserve_queue.pop_front();
			signal_block_reserved.emit(br.block, br.train);
		}
	}
}


template<typename B>
Layout::Storage<B>::~Storage()
{
	while(!base.empty())
		delete *base.begin();
}

template<typename B>
bool Layout::Storage<B>::insert(B &b)
{
	bool result = base.insert(&b).second;
	if(result)
		caches.clear();
	return result;
}

template<typename B>
bool Layout::Storage<B>::erase(B &b)
{
	bool result = base.erase(&b);
	if(result)
		caches.clear();
	return result;
}

template<typename B>
template<typename T>
const set<T *> &Layout::Storage<B>::get() const
{
	for(list<Variant>::const_iterator i=caches.begin(); i!=caches.end(); ++i)
		if(i->check_type<set<T *> >())
			return i->value<set<T *> >();

	caches.push_back(set<T *>());
	set<T *> &result = caches.back().value<set<T *> >();
	for(typename set<B *>::const_iterator i=base.begin(); i!=base.end(); ++i)
		if(T *t = dynamic_cast<T *>(*i))
			result.insert(t);

	return result;
}

template<typename B>
template<typename T>
void Layout::Storage<B>::del()
{
	set<T *> ts = get<T>();
	for(typename set<T *>::iterator i=ts.begin(); i!=ts.end(); ++i)
		if(base.count(*i))
			delete *i;
}


Layout::BlockReservation::BlockReservation(Block &b, Train *t):
	block(b),
	train(t)
{ }


Layout::Loader::Loader(Layout &l):
	DataFile::ObjectLoader<Layout>(l)
{
	add("base",  &Layout::base);
	add("beamgate", &Loader::beamgate);
	add("clock", &Loader::clock);
	add("route", &Loader::route);
	add("signal", &Loader::signal);
	add("terrain", &Loader::terrain);
	add("track", &Loader::track);
	add("train", &Loader::train);
	add("turnout", &Loader::turnout);
	add("zone",  &Loader::zone);

	// Deprecated
	add("turnout", &Loader::turnout2);
}

void Layout::Loader::beamgate()
{
	BeamGate *gate = new BeamGate(obj);
	load_sub(*gate);
}

void Layout::Loader::clock()
{
	load_sub(obj.clock);
}

void Layout::Loader::route()
{
	Route *rte = new Route(obj);
	load_sub(*rte);
}

void Layout::Loader::signal(const string &n)
{
	Signal *sig = new Signal(obj, obj.catalogue.get<SignalType>(n));
	load_sub(*sig);
}

void Layout::Loader::terrain(const string &n)
{
	Terrain *ter = new Terrain(obj, obj.catalogue.get<TerrainType>(n));
	load_sub(*ter);
}

void Layout::Loader::track(const string &n)
{
	Track *trk = new Track(obj, obj.catalogue.get<TrackType>(n));
	load_sub(*trk);
	const set<Track *> &tracks = obj.objects.get<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if(*i!=trk)
			trk->link_to(**i);
}

void Layout::Loader::train(const string &n, unsigned addr, const std::string &proto)
{
	Train *trn = new Train(obj, obj.catalogue.get<VehicleType>(n), addr, proto);
	load_sub(*trn);
}

void Layout::Loader::turnout(unsigned addr)
{
	/* Kinda slow, but maintaining a map of turnouts seems more trouble than
	it's worth at this point. */
	const set<Track *> &tracks = obj.objects.get<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_turnout_address()==addr)
		{
			load_sub(**i);
			return;
		}

	throw key_error(addr);
}

void Layout::Loader::turnout2(unsigned addr, unsigned path)
{
	// XXX Discrepancy between ids and addresses
	if(obj.driver)
		obj.driver->set_turnout(addr, path);
}

void Layout::Loader::zone()
{
	Zone *zne = new Zone(obj);
	load_sub(*zne);
}

} // namespace R2C2
