#ifndef LIBR2C2_LAYOUT_H_
#define LIBR2C2_LAYOUT_H_

#include <deque>
#include <set>
#include <sigc++/sigc++.h>
#include <msp/datafile/objectloader.h>
#include <msp/time/timedelta.h>
#include <msp/time/timestamp.h>
#include "geometry.h"
#include "sensor.h"
#include "clock.h"

namespace R2C2 {

class ArticleNumber;
class BeamGate;
class Block;
class Catalogue;
class Driver;
class Object;
class Route;
class Signal;
class Track;
class TrackChain;
class Train;
class Vehicle;
class Zone;

class Layout
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<Layout>
	{
	public:
		Loader(Layout &);
	private:
		void beamgate();
		void clock();
		void route();
		void signal(const std::string &);
		void terrain(const std::string &);
		void track(const std::string &);
		void train(const std::string &, unsigned, const std::string &);
		void turnout(unsigned);
		void turnout2(unsigned, unsigned);
		void zone();
	};

private:
	template<typename B>
	class Storage
	{
	private:
		std::set<B *> base;
		mutable std::list<Msp::Variant> caches;

	public:
		~Storage();

		bool insert(B &);
		bool erase(B &);

		const std::set<B *> &get() const { return base; }

		template<typename T>
		const std::set<T *> &get() const;

		template<typename T>
		void del();
	};

	struct BlockReservation
	{
		Block &block;
		Train *train;

		BlockReservation(Block &, Train *);
	};

public:
	typedef std::vector<Zone *> ZoneArray;

	sigc::signal<void, Object &> signal_object_added;
	sigc::signal<void, Object &> signal_object_removed;
	sigc::signal<void, TrackChain &> signal_track_chain_added;
	sigc::signal<void, TrackChain &> signal_track_chain_removed;
	sigc::signal<void, Train &> signal_train_added;
	sigc::signal<void, Train &> signal_train_removed;
	sigc::signal<void, Block &, Train *> signal_block_reserved;
	sigc::signal<void, Sensor &, Sensor::State> signal_sensor_state_changed;
	sigc::signal<void, Block *, const std::string &> signal_emergency;

private:
	Catalogue &catalogue;
	Driver *driver;
	Clock clock;
	std::string base;
	Storage<Object> objects;
	Storage<TrackChain> track_chains;
	Storage<Sensor> sensors;
	std::map<unsigned, Train *> trains;
	Msp::Time::TimeStamp last_tick;
	unsigned next_turnout_addr;
	bool emitting_block_reserved;
	std::deque<BlockReservation> block_reserve_queue;

public:
	Layout(Catalogue &, Driver * = 0);
	~Layout();

	Catalogue &get_catalogue() const { return catalogue; }
	bool has_driver() const { return driver; }
	Driver &get_driver() const;
	Clock &get_clock() { return clock; }
	const std::string &get_base() const { return base; }

	void add(Object &);
	void add(Track &);
	void add(TrackChain &);
	void add(Block &);
	void add(Sensor &);
	void add(BeamGate &);

	template<typename T>
	const std::set<T *> &get_all() const;

	template<typename T>
	T *pick(const Ray &);

	void remove(Object &);
	void remove(Track &);
	void remove(TrackChain &);
	void remove(Sensor &);
	void remove(BeamGate &);

	unsigned allocate_turnout_address();

	Block &get_block(unsigned) const;
	void create_blocks();
	void create_blocks(Track &);

	Route &get_route(const std::string &) const;
	void update_routes();

	ZoneArray get_zones(const std::string &) const;
	Zone &get_zone(const std::string &, unsigned) const;

	void add_train(Train &);
	Train &get_train(unsigned) const;
	const std::map<unsigned, Train *> &get_trains() const { return trains; }
	void remove_train(Train &);

	void tick(float = 1.0f);
private:
	void step(const Msp::Time::TimeDelta &);
public:
	void emergency(Block *, const std::string &);

	void save(const std::string &) const;
	void save_dynamic(const std::string &) const;
private:
	void sensor_state_changed(Sensor &, Sensor::State);
	void block_reserved(Block &, Train *);
};

} // namespace R2C2

#endif

